/**
 * Copyright 2019-2021 Britanicus <marcusbritanicus@gmail.com>
 *
 * This file is a part of DesQ Shell (https://gitlab.com/DesQ/Shell)
 * DesQ Shell is the main Shell UI of the DesQ Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * at your option, any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 **/

#include "DesQMenu.hpp"
#include "DesQMenuAdaptor.hpp"

#include "Terminator.hpp"

QDBusInterface session(
    "org.DesQ.Session",
    "/org/DesQ/Session",
    "org.DesQ.Session",
    QDBusConnection::sessionBus()
);

static inline bool canDo( QString action ) {
    QDBusReply<bool> reply = session.call( "Can" + action );

    return reply.value();
}


DesQMenu::DesQMenu() : QWidget() {
    /* Title */
    setWindowTitle( "DesQ Menu" );

    /* No frame, and stay on top */
    setWindowFlags( Qt::Popup | Qt::FramelessWindowHint | Qt::WindowStaysOnTopHint );

    setAttribute( Qt::WA_TranslucentBackground );

    /* DBus Adaptor */
    new DesQMenuAdaptor( this );

    QDBusConnection dbusSess = QDBusConnection::sessionBus();

    dbusSess.registerObject( "/org/DesQ/Menu", this );
    dbusSess.registerService( "org.DesQ.Menu" );
}


void DesQMenu::setup() {
    searchEdit = new QLineEdit();
    searchEdit->setPlaceholderText( "Press '/' to start search..." );
    searchEdit->setFont( QFont( font().family(), 20 ) );
    searchEdit->setAlignment( Qt::AlignCenter );
    searchEdit->setFrame( QFrame::NoFrame );
    searchEdit->setStyleSheet( "background: transparent; margin-bottom: 10px;" );
    searchEdit->setMinimumWidth( 300 );
    searchEdit->setSizePolicy( QSizePolicy( QSizePolicy::MinimumExpanding, QSizePolicy::Preferred ) );

    /** Timer to delay search start */
    delayTimer = new QBasicTimer();

    catsView = new DesQCategoryView( this );
    appsView = new DesQApplicationsView( this );
    favsView = new DesQFavoritesView( this );

    catsView->setCategories( appsView->availableCategories() );

    connect(
        searchEdit, &QLineEdit::textEdited, [ = ] () {
            delayTimer->start( 250, Qt::PreciseTimer, this );
        }
    );

    connect( appsView, &DesQApplicationsView::menuReloaded, [ = ] () {
                 catsView->setCategories( appsView->availableCategories() );
             } );

    connect( catsView, &DesQCategoryView::loadCategory,  appsView, &DesQApplicationsView::loadCategory );
    connect( catsView, &DesQCategoryView::closeMenu,     this,     &DesQMenu::close );
    connect( appsView, &DesQApplicationsView::closeMenu, this,     &DesQMenu::close );
    connect( favsView, &DesQFavoritesView::closeMenu,    this,     &DesQMenu::close );

    connect( catsView, &DesQCategoryView::focusSearch,   [ = ] () {
                 searchEdit->setFocus();
             } );
    connect( appsView, &DesQApplicationsView::focusSearch, [ = ] () {
                 searchEdit->setFocus();
             } );
    connect( favsView, &DesQFavoritesView::focusSearch, [ = ] () {
                 searchEdit->setFocus();
             } );

    QGridLayout *viewLyt = new QGridLayout();

    viewLyt->setContentsMargins( 5, 15, 5, 15 );
    viewLyt->setSpacing( 0 );
    viewLyt->addWidget( catsView, 1, 0 );
    viewLyt->addWidget( appsView, 1, 1 );
    viewLyt->addWidget( favsView, 1, 2 );

    QFrame *menuFrame = new QFrame();

    menuFrame->setLayout( viewLyt );
    menuFrame->setObjectName( "frame" );
    menuFrame->setStyleSheet( "#frame{ border: none; border-top: 1px solid gray;; border-bottom: 1px solid gray; }" );

    QToolButton *lockScreenBtn = new QToolButton();

    lockScreenBtn->setIcon( QIcon::fromTheme( "system-lock-screen" ) );
    lockScreenBtn->setIconSize( QSize( 36, 36 ) );
    connect( lockScreenBtn, &QToolButton::pressed, [ = ] () {
                 QProcess::startDetached( "swaylock", {} );
             } );

    QToolButton *logOutBtn = new QToolButton();

    logOutBtn->setIcon( QIcon::fromTheme( "system-log-out" ) );
    logOutBtn->setIconSize( QSize( 36, 36 ) );
    connect(
        logOutBtn, &QToolButton::pressed, [ = ] () {
            closeRunningApps();
            session.call( "Logout" );
        }
    );

    QToolButton *suspendBtn = new QToolButton();

    suspendBtn->setIcon( QIcon::fromTheme( "system-suspend" ) );
    suspendBtn->setIconSize( QSize( 36, 36 ) );
    connect(
        suspendBtn, &QToolButton::pressed, [ = ] () {
            session.call( "Suspend" );
        }
    );
    suspendBtn->setDisabled( not canDo( "Suspend" ) );

    QToolButton *hibernateBtn = new QToolButton();

    hibernateBtn->setIcon( QIcon::fromTheme( "system-suspend-hibernate" ) );
    hibernateBtn->setIconSize( QSize( 36, 36 ) );
    connect(
        hibernateBtn, &QToolButton::pressed, [ = ] () {
            session.call( "Hibernate" );
        }
    );
    hibernateBtn->setDisabled( not canDo( "Hibernate" ) );

    QToolButton *powerOffBtn = new QToolButton();

    powerOffBtn->setIcon( QIcon::fromTheme( "system-shutdown" ) );
    powerOffBtn->setIconSize( QSize( 36, 36 ) );
    connect(
        powerOffBtn, &QToolButton::pressed, [ = ] () {
            closeRunningApps();
            session.call( "Shutdown" );
        }
    );
    powerOffBtn->setDisabled( not canDo( "Shutdown" ) );

    QToolButton *rebootBtn = new QToolButton();

    rebootBtn->setIcon( QIcon::fromTheme( "system-reboot" ) );
    rebootBtn->setIconSize( QSize( 36, 36 ) );
    connect(
        rebootBtn, &QToolButton::pressed, [ = ] () {
            closeRunningApps();
            session.call( "Reboot" );
        }
    );
    rebootBtn->setDisabled( not canDo( "Reboot" ) );

    QHBoxLayout *powerLyt = new QHBoxLayout();

    powerLyt->addStretch();
    powerLyt->addWidget( lockScreenBtn );
    powerLyt->addWidget( new QLabel( "|" ) );
    powerLyt->addWidget( logOutBtn );
    powerLyt->addWidget( suspendBtn );
    powerLyt->addWidget( hibernateBtn );
    powerLyt->addWidget( new QLabel( "|" ) );
    powerLyt->addWidget( powerOffBtn );
    powerLyt->addWidget( rebootBtn );
    powerLyt->addStretch();

    QGridLayout *menuLyt = new QGridLayout();

    menuLyt->addWidget( searchEdit, 0, 0, 1, 2, Qt::AlignCenter );
    menuLyt->addWidget( menuFrame,  1, 0, 1, 2 );
    menuLyt->addLayout( powerLyt, 2, 0, Qt::AlignCenter );

    setLayout( menuLyt );

    appsView->setFocus();
}


void DesQMenu::toggleVisible() {
    if ( isVisible() ) {
        hide();
    }

    else {
        showMaximized();
    }
}


void DesQMenu::mouseReleaseEvent( QMouseEvent *mEvent ) {
    QWidget::mouseReleaseEvent( mEvent );
    close();
}


void DesQMenu::paintEvent( QPaintEvent *pEvent ) {
    QPainter painter( this );

    painter.fillRect( geometry(), QColor( 0, 0, 0, 230 ) );
    painter.end();

    QWidget::paintEvent( pEvent );
}


void DesQMenu::closeEvent( QCloseEvent *cEvent ) {
    emit closed();

    /** Close the view first */
    QWidget::closeEvent( cEvent );
    qApp->processEvents();

    /** Clear the search edit */
    searchEdit->clear();

    /** Reset the search on close */
    appsView->setFilter( QString() );

    /** Enable al lthe categories */
    catsView->setEnabled( true );

    /** Reset to default */
    catsView->setCurrentItem( catsView->item( 0 ) );
    catsView->loadCategory( "All Applications" );

    /** Clear all selections and focus the Applications view */
    appsView->selectionModel()->clearSelection();
    appsView->setFocus();
}


void DesQMenu::timerEvent( QTimerEvent *tEvent ) {
    if ( delayTimer->timerId() == tEvent->timerId() ) {
        delayTimer->stop();

        QString filter( searchEdit->text() );
        appsView->setFilter( filter );

        if ( filter.isEmpty() ) {
            catsView->setEnabled( true );
        }

        else {
            catsView->setDisabled( true );
        }

        return;
    }

    QWidget::timerEvent( tEvent );
}
